/*
** Module   : BufferList.java
** Abstract : stores the list of all buffers by the same name in a source file
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050922   @22788 First version which provides a helper class
**                           to store the list of all buffers by the same
**                           name in a source file.  This allows the
**                           complete list of buffer scopes to be 
**                           managed during pattern engine processing.
** 002 GES 20060417   @25565 Added zombie list support.
** 003 GES 20090518   @42375 Import change.
** 004 GES 20100819          Use a different map to provide a stable iteration
**                           order.
** 005 SVL 20130213          Added readOnly property.
** 006 SVL 20131115          Added forceDMOAlias property.
*/

package com.goldencode.p2j.convert;

import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.uast.*;

/**
 * This class contains the data and helper methods to implement/manage the
 * buffer scoping process during conversion.  Each instance of a buffer list
 * corresponds with the list of all buffer scopes by that given name.  This
 * class also allows other state to be stored related to this given buffer
 * name.
 */
public class BufferList
{
   /** The Progress qualified buffer name being referenced. */ 
   private String uniqueName = null;
   
   /** The Progress buffer name being referenced. */ 
   private String bufName = null;

   /** Explicit DMO alias for the converted buffer. */
   private String forceDMOAlias = null;
   
   /** The fully qualified Progress schema name being referenced. */
   private String schemaName = null;
   
   /** The explicit Progress buffer's Java name. */
   private String javaName = null;
   
   /** The logical database name for the buffer. */
   private String dbName = null;
   
   /** <code>true</code> if the buffer was not explicitly defined. */
   private boolean implicit = false;

   /**
    * <code>true</code> if the buffer is read-only (like the OLD buffer of a
    * database WRITE trigger).
    */
   private boolean readOnly = false;
   
   /**
    * All <code>BufferScopeTracker</code> instances that use this buffer.
    * Provides a stable iteration order.
    */
   private Set buffers = new LinkedHashSet();
   
   /** A staging area to temporarily hold a buffer until the block opens. */
   private BufferScopeTracker staged = null;
   
   /** A pending free reference for this buffer name. */
   private BufferScopeTracker pending = null;
   
   /** The current active buffer scope for this buffer name. */
   private BufferScopeTracker active = null;
   
   /** All non-scoping free references are stored in this special scope. */
   private BufferScopeTracker zombies = null;
   
   /**
    * Create an instance to track a buffer scope of a given buffer name
    * which is directly mapped to a specific fully-qualified schema name.
    *
    * @param    unique
    *           The buffer's qualified name.
    * @param    buf
    *           The buffer's Progress name.
    * @param    forceDMOAlias
    *           Explicit DMO alias for the converted buffer. Used for dynamic buffers.
    *           <code>null</code> if auto-converted alias should be used.
    * @param    schema
    *           The fully-qualified schema name referenced by this buffer.
    * @param    java
    *           The converted Java name for this buffer. <code>null</code>
    *           if this is an implicitly defined buffer.
    * @param    dbname
    *           The logical database name for the buffer (this may be
    *           different from the database name in the schema parameter).
    * @param    implicit
    *           <code>true</code> if this buffer is being created due to
    *           an implicit record reference, <code>false</code> if the
    *           buffer was explicitly defined (e.g. with a DEFINE BUFFER
    *           statement).
    * @param    readOnly
    *           <code>true</code> if the buffer is read-only (like the OLD
    *           buffer of a database WRITE trigger).
    */
   public BufferList(String  unique,
                     String  buf,
                     String  forceDMOAlias,
                     String  schema,
                     String  java,
                     String  dbname,
                     boolean implicit,
                     boolean readOnly)
   {
      uniqueName         = unique;
      bufName            = buf;
      this.forceDMOAlias = forceDMOAlias;
      schemaName         = schema;
      javaName           = java;
      dbName             = dbname;
      this.implicit      = implicit;
      this.readOnly      = readOnly;
   }
   
   /**
    * Returns the buffer's qualified name.
    *
    * @return   The qualified buffer name.
    */
   public String getUniqueName()
   {
      return uniqueName;
   }
   
   /**
    * Returns the buffer's name.
    *
    * @return   The buffer name.
    */
   public String getBufName()
   {
      return bufName;
   }

   /**
    * Get the explicit DMO alias for the converted buffer.
    *
    * @return  the explicit DMO alias for the converted buffer or <code>null</code>
    *          if this buffer uses auto-converted alias.
    */
   public String getForceDMOAlias()
   {
      return forceDMOAlias;
   }

   /**
    * Returns the fully-qualified schema name which is referenced by this
    * buffer.
    *
    * @return   The schema name referenced by this buffer.
    */
   public String getSchemaName()
   {
      return schemaName;
   }
   
   /**
    * Returns the explicitly defined buffer's Java name which is the
    * converted name for this buffer.
    *
    * @return   The Java name for this buffer or <code>null</code> if this
    *           is not an explicitly defined buffer.
    */
   public String getJavaName()
   {
      return javaName;
   }
   
   /**
    * Returns the logical database name for the buffer.  This may be
    * different from the database portion of the fully qualified schema
    * name.
    *
    * @return   The logical database for this buffer.
    */
   public String getDbName()
   {
      return dbName;
   }
   
   /**
    * Specifies if the buffer was created due to an implicit record reference 
    * or due to an explicit definition (e.g. DEFINE BUFFER).
    *
    * @return   <code>true</code> if the buffer was created due to an 
    *           implicit record reference, <code>false</code> otherwise.
    */
   public boolean getImplicit()
   {
      return implicit;
   }

   /**
    * Specifies if the buffer is read-only (like the OLD buffer of a database
    * WRITE trigger).
    *
    * @return  <code>true</code> if the buffer is read-only.
    */
   public boolean isReadOnly()
   {
      return readOnly;
   }

   /**
    * Accesses the current staged strong or weak reference for this buffer
    * name.
    *
    * @return   The staged strong or weak reference or <code>null</code> if
    *           there is no staged reference.
    */
   public BufferScopeTracker getStaged()
   {
      return staged;
   }
   
   /**
    * Sets the current staged strong or weak reference for this buffer
    * name.
    *
    * @param    staged
    *           The strong or weak reference to stage or <code>null</code> to
    *           clear the staged reference.
    */
   public void setStaged(BufferScopeTracker staged)
   {
      this.staged = staged;
   }
   
   /**
    * Accesses the current pending (unbound) free reference for this buffer
    * name.
    *
    * @return   The pending unbound free reference or <code>null</code> if
    *           there is no pending free reference.
    */
   public BufferScopeTracker getPending()
   {
      return pending;
   }
   
   /**
    * Sets the current pending (unbound) free reference for this buffer
    * name.
    *
    * @param    pending
    *           The pending unbound free reference or <code>null</code> to
    *           clear the pending free reference.
    */
   public void setPending(BufferScopeTracker pending)
   {
      this.pending = pending;
   }
   
   /**
    * Accesses the current active buffer scope for this buffer name.
    *
    * @return   The active buffer scope or <code>null</code> if there is no
    *           active buffer scope.
    */
   public BufferScopeTracker getActive()
   {
      return active;
   }
   
   /**
    * Sets the current active buffer scope for this buffer name.
    *
    * @param    active
    *           The active buffer scope or <code>null</code> to
    *           clear the current active scope.
    */
   public void setActive(BufferScopeTracker active)
   {
      this.active = active;
   }   
   
   /**
    * Accesses the list of zombie nodes for this buffer name.  A zombie node
    * is an AST that is marked as a <code>NO_REFERENCE</code> so that it
    * has no affect on buffer scoping calcuations BUT it is still tracked
    * and assigned a buffer such that the references can be resolved at
    * runtime since such references really do access data.
    *
    * @return   The list of zombie nodes or <code>null</code> if there were
    *           no such nodes for this buffer.
    */
   public BufferScopeTracker getZombies()
   {
      return zombies;
   }
   
   /**
    * Add a new buffer scope to the list of all buffer scopes for the
    * current file.
    *
    * @param    bst
    *           The scope to add.
    */
   public void addBufferScope(BufferScopeTracker bst)
   {
      buffers.add(bst);
   }
   
   /**
    * Remove a buffer scope from the list of all buffer scopes for the
    * current file.
    *
    * @param    bst
    *           The scope to remove.
    */
   public void removeBufferScope(BufferScopeTracker bst)
   {
      buffers.remove(bst);
   }   
   
   /**
    * Returns an iterator to the list of buffer scopes associated with this
    * buffer name.
    *
    * @return   An iterator to all {@link BufferScopeTracker} instances
    *           associated with this buffer name.
    */
   public Iterator getScopeList()
   {
      return buffers.iterator();
   }   
   
   /**
    * Add a node to the zombie list, creating the zombie list first if needed.
    * <p>
    * A zombie node is an AST that is marked as a <code>NO_REFERENCE</code>
    * so that it has no affect on buffer scoping calcuations BUT it is still
    * tracked and assigned a buffer such that the references can be resolved
    * at runtime since such references really do access data.
    *
    * @param    node
    *           The zombie node to add.
    */
   public void addZombie(Aast node)
   {
      if (zombies == null)
      {
         int type = ProgressParserTokenTypes.FREE_REFERENCE;
         zombies = new BufferScopeTracker(type);
      }
      
      zombies.addReference(node);
   }   
}
