/*
** Module   : DynamicQueryHelper.java
** Abstract : Helper class for dynamic queries.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 CA  20130929 Created initial version.
** 002 SVL 20131115 Add "foce_force_dmo_alias" annotations for dynamic buffers.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.concurrent.atomic.*;
import java.util.logging.*;

import org.hibernate.mapping.Column;
import org.hibernate.mapping.PersistentClass;
import org.hibernate.mapping.Property;

import antlr.*;

import com.goldencode.ast.*;
import com.goldencode.compile.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;

/**
 * Helper class for dynamic query conversion.
 */
class DynamicQueryHelper
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(DynamicQueryHelper.class);
   
   /** The package for the in-mem compiled query classes. */
   public static final String IN_MEM_QUERY_BASE_PAKAGE = "com.goldencode.p2j.persist.dynquery";
   
   /** Map of P2J wrappers to their associated token type. */
   private static final Map<String, Integer> wrapperToTokenType = new HashMap<String, Integer>();
   
   /**
    * Reference to an instance of {@link InMemoryClassLoader} that presents in the chain
    * of system class loaders. If it is not present, the value is <code>null</code>.
    */
   private static final InMemoryClassLoader systemInMemoryClassLoader;
   
   /** A counter of compiled dynamic query classes. */
   private static final AtomicLong queryCounter = new AtomicLong(1);
   
   /** Reference to context-local data. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         
         TransactionManager.registerFinalizable(wa, true);
         
         return wa;
      };
   };
   
   /** Map of the pristine ASTs (.dict and .p2o) for the permanent DBs. */
   private static Map<String, Aast> pristineDbAsts = null;
   
   static
   {
      wrapperToTokenType.put(ParmType.CHAR.toString(),   ProgressParserTokenTypes.FIELD_CHAR);
      wrapperToTokenType.put(ParmType.INT.toString(),    ProgressParserTokenTypes.FIELD_INT);
      wrapperToTokenType.put(ParmType.INT64.toString(),  ProgressParserTokenTypes.FIELD_INT64);
      wrapperToTokenType.put(ParmType.DEC.toString(),    ProgressParserTokenTypes.FIELD_DEC);
      wrapperToTokenType.put(ParmType.LOG.toString(),    ProgressParserTokenTypes.FIELD_LOGICAL);
      wrapperToTokenType.put(ParmType.DATE.toString(),   ProgressParserTokenTypes.FIELD_DATE);
      wrapperToTokenType.put(ParmType.DT.toString(),     ProgressParserTokenTypes.FIELD_DATETIME);
      wrapperToTokenType.put(ParmType.DTTZ.toString(),   ProgressParserTokenTypes.FIELD_DATETIME_TZ);
      wrapperToTokenType.put(ParmType.HANDLE.toString(), ProgressParserTokenTypes.FIELD_HANDLE);
      wrapperToTokenType.put(ParmType.RECID.toString(),  ProgressParserTokenTypes.FIELD_RECID);
      wrapperToTokenType.put(ParmType.ROWID.toString(),  ProgressParserTokenTypes.FIELD_ROWID);

      // check the class-loader
      systemInMemoryClassLoader = InMemoryClassLoader.getSystemInMemoryClassLoader();
      if (systemInMemoryClassLoader == null)
      {
          throw new IllegalStateException("It is expected to have InMemoryClassLoader " +
                                          "in the chain of system class loaders.");
      }

      // check the ast manager plugin
      if (!AstManager.isPluginOfType(InMemoryRegistryPlugin.class))
      {
         throw new IllegalStateException(
            "The InMemoryRegistryPlugin must be used to parse dynamic queries!");
      }
   }
   
   /**
    * Parse the given query string and return a {@link P2JQuery} instance.
    * 
    * @param    buffers
    *           The list of buffers used by the query.
    * @param    predicate
    *           The query predicate.
    * @param    qname
    *           The query's name.
    * 
    * @return   A {@link P2JQuery} instance representing the given legacy query string.
    */
   static P2JQuery parseQuery(List<Buffer> buffers, String predicate, String qname)
   {
      WorkArea wa = locate();

      loadPermanentSchemas();

      if (!wa.permSchemasLoaded)
      {
         return null;
      }
      
      long queryId = queryCounter.getAndIncrement();

      // set the info about what we are currently parsing
      wa.predicate     = predicate;
      wa.qname         = qname;
      wa.buffers       = buffers;
      wa.dynQueryClass = String.format("InMemP2jQuery%d", queryId);
      wa.fullClassName = String.format("%s.%s", IN_MEM_QUERY_BASE_PAKAGE, wa.dynQueryClass);
      wa.dynQueryFile  = String.format(".%sin-mem-p2j-query%d.p", File.separator, queryId);
      wa.jastFile      = String.format("%s.jast", wa.dynQueryFile);
      wa.astFile       = String.format("%s.ast", wa.dynQueryFile);
      wa.p2oFile       = String.format("%s.p2o", wa.dynQueryFile);

      try
      {
         StringBuilder         pcode = new StringBuilder();
         SymbolResolver        sym   = null;
         ByteArrayOutputStream code  = null;
   
         // step 1: build the symbol resolver and load the schema
         sym = new SymbolResolver();
   
         // set the schema dictionary used by this context
         SchemaDictionary dict = SchemaDictionary.localInstance(); 
         sym.setSchemaDictionary(dict);
   
         try
         {
            // step 2: load p2o and schema definitions for all temporary buffers
   
            // add a default scope that represents the scope of the external procedure
            sym.addSchemaScope(false);
            
            // load the p2o and schema data
            try
            {
               prepareTempTable(dict, pcode);
            }
            catch (PersistenceException exc)
            {
               showConversionError(exc, "prepare the temp-table buffers for");

               if (LOG.isLoggable(Level.SEVERE))
               {
                  String msg = "Error preparing the temp-table buffers for predicate '%s'";
                  LOG.log(Level.SEVERE, String.format(msg, predicate));
               }
               
               return null;
            }
   
            // step 3: parse the predicate
            pcode.append(String.format("OPEN QUERY inMemQuery%d %s.", queryId, predicate));
   
            boolean        complete = false;
            Exception      err      = null;
            StringReader   in       = new StringReader(pcode.toString());
            ProgressLexer  lexer    = new ProgressLexer(in, sym);
            ProgressParser parser   = new ProgressParser(lexer, sym);
            try
            {
               complete = parser.external_proc();
            }
            catch (RecognitionException | TokenStreamException exc)
            {
               err = exc;
               complete = false;
            }
   
            if (!complete)
            {
               showConversionError(err, "parse");
               return null;
            }
         
            try
            {
               ProgressAst result = (ProgressAst) parser.getAST();
               result.brainwash(wa.dynQueryFile, true);
               
               // step 4: prepare the initial AST
               prepareQueryTree(result);
               
               // step 5: run conversion
               runPattern(result, "annotations/annotations", true);
               result = (ProgressAst) AstManager.get().loadTree(wa.astFile);
               
               runPattern(result, "convert/base_structure", true);
               result = (ProgressAst) AstManager.get().loadTree(wa.astFile);
               
               runPattern(result, "convert/core_conversion", true);
               result = (ProgressAst) AstManager.get().loadTree(wa.astFile);
               
               // walk the AST and collect all the buffers
               List<String> realBuffers = collectBuffers(result);

               // validate the buffers
               if (!validateBuffers(realBuffers))
               {
                  return null;
               }
               
               JavaAst jcode = (JavaAst) AstManager.get().loadTree(wa.jastFile);

               // adjust the java AST so that it follows these rules:
               // 1. all fields defining a buffer are public and initialized to null
               // 2. the query variable is defined by an instance field
               // 3. all code in the "execute" method is changed so:
               //    - the anon Block instance is removed and the code moved to the "execute" method
               //    - the "RecordBuffer.openScope" calls are removed
               //    - the "query.open()" call is removed
               // 4. a "getQuery" getter for the query at step 2 is added

               // generate the java code
               PatternEngine engine = null;
               
               // following brewing is for debug only
               /*
               engine = runPattern(jcode, "convert/brew", false);
               code = (ByteArrayOutputStream) engine.getStoredObject(wa.dynQueryClass);
               System.out.println(code);
               */

               runPattern(jcode, "convert/fix_in_mem_query", false);
               Aast qast = (Aast) AstManager.get().loadTree(wa.jastFile);
         
               // generate the java code
               engine = runPattern(qast, "convert/brew", false);
               code = (ByteArrayOutputStream) engine.getStoredObject(wa.dynQueryClass);
            }
            catch (Exception exc)
            {
               showConversionError(exc, "convert");
               return null;
            }
         }
         finally
         {
            // cleanup the schema
            sym.deleteSchemaScope(false);
            
            // cleanup the progress and java ASTs, they are no longer needed
            AstManager.get().removeTree(wa.astFile);
            AstManager.get().removeTree(wa.jastFile);
         }
   
         // step 6: in-memory compile of the generated ast
         P2JQuery p2jQuery = compileQuery(code);
         
         return p2jQuery;
      }
      finally
      {
         // reset the WorkArea temporary data
         wa.predicate     = null;
         wa.qname         = null;
         wa.buffers       = null;
         wa.dynQueryClass = null;
         wa.fullClassName = null;
         wa.dynQueryFile  = null;
         wa.jastFile      = null;
         wa.astFile       = null;
         wa.p2oFile       = null;
      }
   }
   
   /**
    * Called when a query is deleted. It unregisters its associated class, in case it was a
    * dynamically created query.
    * 
    * @param    query
    *           The query which needs to have its query class deregistered from the class-loader.
    */
   static void unregisterQuery(P2JQuery query)
   {
      Class<?> qclass = locate().dynamicQueries.remove(query);

      if (qclass != null)
      {
         systemInMemoryClassLoader.deregister(qclass.getName());
      }
   }
   
   /**
    * Process the definitions of the temp-tables (associated with the temporary buffers in the given
    * list) to create the associated dictionary and p2o ASTs.
    * <p>
    * Also, it appends to the progress code string any needed 4GL code (like DEFINE BUFFER 
    * statements for explicit buffers, to refer back the originating table or CREATE ALIAS 
    * statements in case of explicit aliases active for this context).
    * 
    * @param    dict
    *           The schema dictionary where the temp-tables will be loaded.
    * @param    pcode
    *           Append any new progress code here.
    * 
    * @throws   PersistenceException
    *           In case of problems during preparation of the temp-table schema and p2o ASTs.
    */
   private static void prepareTempTable(SchemaDictionary dict, StringBuilder pcode)
   throws PersistenceException
   {
      WorkArea wa = locate();
      
      DataModelAst p2oRoot = buildRootDataModel(DatabaseManager.TEMP_TABLE_SCHEMA);
      p2oRoot.brainwash(wa.p2oFile);
      ProgressAst schemaRoot = buildRootSchema(DatabaseManager.TEMP_TABLE_SCHEMA);

      integer tableOrder = new integer(0);
      Set<String> ldbs = new TreeSet<String>();
      for (Buffer buf : wa.buffers)
      {
         BufferImpl impl = (BufferImpl) buf;
         RecordBuffer rbuf = impl.buffer();
         rbuf.initialize();
         ldbs.add(rbuf.getLogicalDatabase().toLowerCase());

         String legacyBufName = impl.getName().toStringMessage();
         String legacyTableName = TableMapper.getLegacyName(buf);
         
         if (rbuf.isTemporary())
         {
            // create the schema and p2o definition for this temp-table
            DataModelAst model = buildDataModelClass(buf, rbuf);
            ProgressAst table = buildTempTable(model, rbuf, tableOrder);

            schemaRoot.graft(table);
            
            p2oRoot.graft(model);
         }
         
         // must create explicit buffers, if the legacy buffer name is not the same as the legacy
         // default buffer name (which is the same as the legacy table name).
         if (!legacyBufName.equalsIgnoreCase(legacyTableName))
         {
            String tempTable = rbuf.isTemporary() ? "TEMP-TABLE" : ""; 
            
            String defBufStmt = String.format("DEFINE BUFFER %s FOR %s %s.%n", 
                                              legacyBufName, tempTable, legacyTableName);
            pcode.append(defBufStmt);
         }
      }
      
      // custom database aliases...
      Map<String, String> aliasesByLDB = ConnectionManager.get().getAliasesByLDB();
      
      for (String ldb : ldbs)
      {
         String alias = aliasesByLDB.get(ldb.toLowerCase());
         if (alias != null)
         {
            String stmt = "CREATE ALIAS %s FOR DATABASE %s.%n";
            
            pcode.append(String.format(stmt, alias, ldb));
         }
      }

      // save the P2O tree, to expose it to the conversion rules
      AstManager.get().saveTree(p2oRoot, wa.p2oFile, true);

      // load the temp schema
      dict.loadFromAst(schemaRoot);
   }
   
   /**
    * Load the permanent schema ASTs into the current context's AST manager plugin.
    * <p>
    * Each context will actually execute this method once. On the very first call (regardless of 
    * context), it will load the ASTs from the file-system into the {@link #pristineDbAsts} map.
    */
   private static void loadPermanentSchemas()
   {
      if (locate().permSchemasLoaded)
      {
         return;
      }
      
      SchemaConfig sc = Configuration.getSchemaConfig();
      Iterator<String> dbIter = sc.defaultDatabases();
      List<String> all = new ArrayList<String>();
      while (dbIter.hasNext())
      {
         all.add(dbIter.next());
      }
      all.add(sc.getMetadata().getName());
   
      // load the schema dict files into the user's in-memory AST storage
      synchronized (DynamicQueryHelper.class)
      {
         if (pristineDbAsts == null)
         {
            pristineDbAsts = new HashMap<>();
            String name = Configuration.getParameter("registry");
            XmlFilePlugin fileLoader = new XmlFilePlugin(name);
            
            for (String schema : all)
            {
               try
               {
                  String schemaPath = sc.getSchemaFile(schema).getPath();
                  schemaPath = Configuration.normalizeFilename(schemaPath);

                  // load the schema
                  if (fileLoader.isExistingAst(schemaPath))
                  {
                     Aast ast = fileLoader.loadTree(schemaPath);
                     pristineDbAsts.put(schemaPath, ast);
                  }
      
                  // load the .p2o definitions
                  String p2oPath = schemaPath.replaceAll("\\.dict", ".p2o");
                  p2oPath = Configuration.normalizeFilename(p2oPath);
                  if (fileLoader.isExistingAst(p2oPath))
                  {
                     Aast ast = fileLoader.loadTree(p2oPath);
                     pristineDbAsts.put(p2oPath, ast);
                  }
               }
               catch (AstException | ConfigurationException exc)
               {
                  pristineDbAsts = null;

                  String msg = String.format("Could not load the pristine %s schema!", schema);

                  if (LOG.isLoggable(Level.SEVERE))
                  {
                     LOG.log(Level.SEVERE, msg, exc);
                  }
                  
                  throw new RuntimeException(msg, exc);
               }
            }
         }
      }
      
      for (String schema : all)
      {
         try
         {
            String schemaPath = sc.getSchemaFile(schema).getPath();
            schemaPath = Configuration.normalizeFilename(schemaPath);

            // load the schema
            if (pristineDbAsts.containsKey(schemaPath))
            {
               Aast schemaAst = pristineDbAsts.get(schemaPath);
               AstManager.get().saveTree(schemaAst, schemaPath, false);
            }
      
            // load the .p2o definitions
            String p2oPath = schemaPath.replaceAll("\\.dict", ".p2o");
            p2oPath = Configuration.normalizeFilename(p2oPath);
            if (pristineDbAsts.containsKey(p2oPath))
            {
               Aast p2oAst = pristineDbAsts.get(p2oPath);
               AstManager.get().saveTree(p2oAst, p2oPath, false);
            }
         }
         catch (ConfigurationException exc)
         {
            String msg = String.format("Could not load the context-local %s schema!", schema);
            
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.log(Level.SEVERE, msg, exc);
            }
            
            throw new RuntimeException(msg, exc);
         }
      }
   
      locate().permSchemasLoaded = true;
   }

   /**
    * Prepare the query AST by placing the appropriate annotations with the converted java names
    * for the external program and also by created bogus
    * {@link ProgressParserTokenTypes#DEFINE_TEMP_TABLE} nodes (which would be created from a
    * <code>DEFINE TEMP-TABLE tt.</code> statement, with no fields); the later is needed to allow
    * proper conversion of the default buffer name.
    * 
    * @param    root
    *           The root AST of the converted progress code.
    */
   private static void prepareQueryTree(AnnotatedAst root)
   {
      WorkArea wa = locate();

      // Create a bogus DEFINE TEMP-TABLE for each temp buffer (just to let their converted
      // buffer names resolve properly).
      // Dynamic buffers have unique DMO aliases, so "some-buffer" name may be converted to the
      // alias like "someTable__1", while aliases in predicate will be converted using standard
      // P2J rules, which will result to the converted alias "someTable". So below we explicitly
      // specify converted DMO alias like "someTable__1" for the conversion rules using
      // "force_dmo_alias" annotation.
      for (Buffer buf : wa.buffers)
      {
         BufferImpl impl = (BufferImpl) buf;
         String legacyTableName = TableMapper.getLegacyName(buf);
         
         if (!impl.buffer().isTemporary())
         {
            continue;
         }
         
         ProgressAst stmt = createProgressAst(ProgressParserTokenTypes.STATEMENT, "statement");
         stmt.putAnnotation("reachable", true);
         
         ProgressAst defChild = createProgressAst(ProgressParserTokenTypes.DEFINE_TEMP_TABLE, 
                                                  "DEFINE");
         defChild.putAnnotation("reachable", true);
         defChild.putAnnotation("schemaname", legacyTableName);
         defChild.putAnnotation("bufname", legacyTableName);
         defChild.putAnnotation("dbname", "");
         if (impl.buffer().isDynamic())
            defChild.putAnnotation("force_dmo_alias", impl.buffer().getDMOAlias());
         stmt.graft(defChild);
         
         ProgressAst ttChild = createProgressAst(ProgressParserTokenTypes.KW_TEMP_TAB, 
                                                 "TEMP-TABLE");
         ttChild.putAnnotation("reachable", true);
         defChild.graft(ttChild);
         
         ProgressAst symChild = createProgressAst(ProgressParserTokenTypes.SYMBOL,
                                                  legacyTableName);
         symChild.putAnnotation("reachable", true);
         ttChild.graft(symChild);

         root.graftAt(stmt, 0);
      }
      
      // put annotation related to name, package, etc
      root.putAnnotation("file", wa.dynQueryFile);
      root.putAnnotation("classname", wa.dynQueryClass);
      root.putAnnotation("javaname", wa.dynQueryClass);
      root.putAnnotation("relative-name", wa.dynQueryFile);
      root.putAnnotation("basename", wa.dynQueryFile);
      root.putAnnotation("package-base", "");
      root.putAnnotation("pkgname", IN_MEM_QUERY_BASE_PAKAGE);

      // make "force_dmo_alias" annotations, errors are ignored at this point, they will
      // be raised later anyway
      Iterator<ProgressAst> iter = root.iterator();
      while (iter.hasNext())
      {
         ProgressAst chAst = iter.next();

         if (chAst.getType() == ProgressParserTokenTypes.RECORD_PHRASE ||
             chAst.getType() == ProgressParserTokenTypes.DEFINE_BUFFER)
         {
            ProgressAst bufAst = (ProgressAst) chAst.getFirstChild();
            String bufName = (String) bufAst.getAnnotation("bufname");
            if (bufName != null)
            {
               int dotPos = bufName.lastIndexOf(".");
               if (dotPos >= 0)
                  bufName = bufName.substring(dotPos + 1);

               int idx = locateBuffer(bufName);
               if (idx >= 0)
               {
                  BufferImpl impl = (BufferImpl) wa.buffers.get(idx);
                  if (impl.buffer().isDynamic())
                     bufAst.putAnnotation("force_dmo_alias", impl.buffer().getDMOAlias());
               }
            }
         }
      }
   }
   
   /**
    * Get the buffer names used by the AST produced for converted predicated.
    * 
    * @param    result
    *           The list of buffer names.
    *           
    * @return   See above.
    */
   private static List<String> collectBuffers(ProgressAst result)
   {
      List<String> buffers = new ArrayList<>();
      
      Iterator<ProgressAst> iter = result.iterator();
      while (iter.hasNext())
      {
         ProgressAst chAst = iter.next();
         
         if (chAst.getType() != ProgressParserTokenTypes.RECORD_PHRASE)
         {
            continue;
         }
         
         ProgressAst bufAst = (ProgressAst) chAst.getFirstChild();
         String bufname = (String) bufAst.getAnnotation("bufname");
         
         if (bufname.indexOf(".") >= 0)
         {
            bufname = bufname.substring(bufname.lastIndexOf(".") + 1);
         }
         
         buffers.add(bufname);
      }
      
      return buffers;
   }
   
   /**
    * Validate the list of buffers against the buffers loaded by this QUERY resource (loaded in
    * the {@link WorkArea#buffers} list.
    * <p>
    * The buffers must have (all conditions must be true):
    * <ol>
    *    <li>The buffers must be in the same order as at the QUERY resource.</li>
    *    <li>All the buffers loaded by this QUERY resource must be referenced.</li>
    *    <li>No buffer outside the QUERY's buffer list is allowed.</li>
    * </ol>
    * 
    * @param    realBuffers
    *           The list of buffer names used at the query predicated.
    *           
    * @return   <code>true</code> if the buffers are valid.
    */
   private static boolean validateBuffers(List<String> realBuffers)
   {
      WorkArea wa = locate();

      // check if the expected and converted buffer list match
      if (realBuffers.size() != wa.buffers.size())
      {
         final String errMsg =
            "QUERY-PREPARE text must have 1 FOR EACH/PRESELECT for each query buffer";

         ErrorManager.recordOrShowError(7325, errMsg, true, false, false);

         return false;
      }

      // check buffer positions and if known buffer
      for (int i = 0; i < realBuffers.size(); i++)
      {
         String realBuffer = realBuffers.get(i);

         int idx = locateBuffer(realBuffer);
         if (idx == -1)
         {
            String errMsg = " %s must be an unabbreviated name of a buffer known in query %s";
            errMsg = String.format(errMsg, realBuffer, wa.qname);
            
            ErrorManager.recordOrShowError(7327, errMsg, true, false, false);
            
            return false;
         }
         
         if (idx != i)
         {
            String errMsg = "Buffer %s is not referenced in the same order in PREPARE as " +
                            "in BUFFERS list for query %s";
            errMsg = String.format(errMsg, realBuffer, wa.qname);
            
            ErrorManager.recordOrShowError(7326, errMsg, true, false, false);

            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Locate a {@link Buffer} with the given <code>name</code>, from the {@link WorkArea#buffers}
    * list and return its position.
    * 
    * @param    name
    *           The buffer name.
    *           
    * @return   The 0-based position of this buffer or -1 if not found.
    */
   private static int locateBuffer(String name)
   {
      WorkArea wa = locate();
      
      int idx = 0;
      for (Buffer buf : wa.buffers)
      {
         if (name.equalsIgnoreCase(((CommonHandleChain) buf).getName().toStringMessage()))
         {
            return idx;
         }
         
         idx++;
      }
      
      return -1;
   }
   
   /**
    * Compile the given java code and extract a {@link P2JQuery} instance.
    * 
    * @param    code
    *           The code to be compiled.
    *           
    * @return   A {@link P2JQuery} instance or <code>null</code> in case of compile or other kind
    *           of problems.
    */
   private static P2JQuery compileQuery(ByteArrayOutputStream code)
   {
      WorkArea wa = locate();
      P2JQuery query = null;

      // compile the code
      StringBuilder errs = new StringBuilder();
      Map<String, Class<?>> classes = InMemoryCompiler.compile(new String[]
                                                               {
                                                                  wa.fullClassName,
                                                               },
                                                               new String[]
                                                               {
                                                                  code.toString()
                                                               },
                                                               errs, true);

      if (classes == null)
      {
         errs.append('\n').append(code.toString());
         showConversionError(new Exception(errs.toString()), "compile the java code for");
         return null;
      }
      
      Class<?> qclass = classes.get(wa.fullClassName);
      
      try
      {
         Object qinst = qclass.newInstance();

         // set the buffer fields
         for (Buffer buf : wa.buffers)
         {
            RecordBuffer rbuf = ((BufferImpl) buf).buffer();
            
            String bufName = rbuf.getDMOAlias();
            
            try
            {
               Field fb = qclass.getField(bufName);
               fb.set(qinst, fb.getType().cast(buf));
            }
            catch (NoSuchFieldException exc)
            {
               String msg = String.format("assign the %s P2J buffer for", bufName);
               showConversionError(exc, msg);

               return null;
            }
         }
         
         // no open scope calls are needed for the buffers, as they inherit the caller
         
         // invoke "execute" method to create the P2J query; the silent mode needs to be backed
         // up and restored, as it might get overridden.
         boolean silent = ErrorManager.isSilentError();
         try
         {
            Method mexecute = qclass.getMethod("execute");
            mexecute.invoke(qinst);
         }
         finally
         {
            ErrorManager.setSilent(silent);
         }

         // get the P2JQuery instance
         Method mget = qclass.getMethod("getQuery");
         query = (P2JQuery) mget.invoke(qinst);
         
         // this class needs to survive until the query is deleted
         locate().dynamicQueries.put(query, qclass);
      }
      catch (InstantiationException   |
             IllegalAccessException   |
             SecurityException        |
             NoSuchMethodException    |
             IllegalArgumentException | 
             InvocationTargetException exc)
      {
         showConversionError(exc, "instantiate the P2JQuery for");
         return null;
      }

      return query;
   }
   
   /**
    * Run the given pipeline in a dedicated {@link PatternEngine engine}.  When the conditional
    * flag is enabled, it will include only the runtime query conversion code.
    * 
    * @param    ast
    *           The AST to use as input source.
    * @param    profile
    *           Name of the rule-set to be ran.
    * @param    conditional
    *           Flag indicating if conditional rule-set processing for dynamic query conversion
    *           is to be enabled.
    *           
    * @return   The {@link PatternEngine} instance.
    * 
    * @throws   ConfigurationException
    *           if any error occurs loading the specified configuration profile.
    * @throws   AstException
    *           if any error occurs loading a persisted AST.
    */
   private static PatternEngine runPattern(Aast ast, String profile, boolean conditional)
   throws ConfigurationException,
          AstException
   {
      PatternEngine engine = new PatternEngine();
      if (conditional)
      {
         engine.setConditionalRuleSets(true);
         engine.addConditionalFlag(ConversionFlag.RUNTIME_QUERY.toString());
      }
      engine.addTargetAst(ast);
      engine.run(profile);
      
      return engine;
   }
   
   /**
    * Build a temp-table definition to be loaded in the {@link SchemaDictionary}. Use the given
    * {@link DataModelAst data model object} AST as the source.
    * 
    * @param    p2o
    *           The data model object AST.
    * @param    rbuf
    *           The record buffer instance.
    * @param    order
    *           The order of this temp-table AST.
    *           
    * @return   See above.
    */
   private static ProgressAst buildTempTable(DataModelAst p2o, RecordBuffer rbuf, integer order)
   {
      String schema      = rbuf.getSchema();
      Class<?> iface     = rbuf.getDMOInterface();
      String legacyTable = (String) p2o.getAnnotation("historical");
      
      ProgressAst root = createProgressAst(SchemaParserTokenTypes.TEMP_TABLE, legacyTable);
      root.putAnnotation("lastOrder", order.toJavaLongType());
      
      int fieldIdx = 0;
      for (int i = 0; i < p2o.getNumImmediateChildren(); i++)
      {
         DataModelAst child = (DataModelAst) p2o.getChildAt(i);
         
         if (child.getType() == DataModelTokenTypes.PROPERTY)
         {
            String datatype = (String) child.getAnnotation("datatype");
            int ftype = wrapperToTokenType.get(datatype);
            String legacyField = (String) child.getAnnotation("historical");

            ProgressAst field = createProgressAst(ftype, legacyField);
            
            // the following properties are not needed by SchemaDictionary:
            // 1. sqlWidth
            // 2. extent
            // 3. decimals
            // 4. length
            // 5. maxwidth
            
            if (DMOIndex.isCaseSensitive(schema, iface, child.getText()))
            {
               field.addChild(createProgressAst(SchemaParserTokenTypes.KW_CASE_SEN, 
                                                "case-sensitive"));
            }

            // add ORDER node
            ProgressAst pord = createProgressAst(SchemaParserTokenTypes.ORDER, "ORDER");
            pord.addChild(createProgressAst(SchemaParserTokenTypes.NUM_LITERAL,
                                            Long.toString(fieldIdx)));
            fieldIdx = fieldIdx + 10;
            
            field.addChild(pord);
            
            Long extent = (Long) child.getAnnotation("extent");
            if (extent != null)
            {
               ProgressAst pext = createProgressAst(SchemaParserTokenTypes.KW_EXTENT, "extent");
               pext.addChild(createProgressAst(SchemaParserTokenTypes.NUM_LITERAL, 
                                               extent.toString()));
               
               field.addChild(pext);
            }
            
            root.addChild(field);
         }
         else if (child.getType() == DataModelTokenTypes.INDEX)
         {
            // add a new index
            String legacyIndex = (String) child.getAnnotation("historical");
            ProgressAst rootIdx = createProgressAst(SchemaParserTokenTypes.INDEX, legacyIndex);
            
            // build index properties
            ProgressAst props = createProgressAst(SchemaParserTokenTypes.PROPERTIES, "");
            if (child.isAnnotation(P2OLookup.UNIQUE))
            {
               props.addChild(createProgressAst(SchemaParserTokenTypes.KW_UNIQUE, "UNIQUE"));
            }
            if (child.isAnnotation(P2OLookup.PRIMARY))
            {
               props.addChild(createProgressAst(SchemaParserTokenTypes.KW_PRIMARY, "PRIMARY"));
            }
            if (child.isAnnotation(P2OLookup.WORD))
            {
               props.addChild(createProgressAst(SchemaParserTokenTypes.KW_WORD_IDX, "WORD"));
            }
            
            rootIdx.addChild(props);
            
            // build index components
            DataModelAst idxCol =
               (DataModelAst) child.getImmediateChild(DataModelTokenTypes.INDEX_COL, null);
            while (idxCol != null)
            {
               String legacyField = (String) child.getAnnotation("historical");
               
               ProgressAst idxField = createProgressAst(SchemaParserTokenTypes.INDEX_FIELD,
                                                        legacyField);
               if (idxCol.isAnnotation(P2OLookup.ABBREVIATED))
               {
                  idxField.addChild(createProgressAst(SchemaParserTokenTypes.KW_ABBV,
                                                      "ABBREVIATED"));
               }
               if (idxCol.isAnnotation(P2OLookup.DESCEND))
               {
                  idxField.addChild(createProgressAst(SchemaParserTokenTypes.KW_DESCEND,
                                                      "DESCENDING"));
               }
               else
               {
                  idxField.addChild(createProgressAst(SchemaParserTokenTypes.KW_ASCEND,
                                                      "ASCENDING"));
               }
               
               rootIdx.addChild(idxField);
               
               idxCol = (DataModelAst) child.getImmediateChild(DataModelTokenTypes.INDEX_COL,
                                                               idxCol);
            }
            
            root.addChild(rootIdx);
         }
      }
      
      order.assign(order.intValue() + 10);
      
      return root;
   }
   
   /**
    * Build the root {@link ProgressAst} node for the schema dictionary AST.
    * 
    * @param    schema
    *           The schema name.
    *           
    * @return   See above.
    */
   private static ProgressAst buildRootSchema(String schema)
   {
      ProgressAst root = createProgressAst(SchemaParserTokenTypes.DATABASE, schema);
      root.putAnnotation("srcfile", "in-memory");
      
      return root;
   }
   
   /**
    * Build the {@link DataModelAst root} AST and associated with the given schema.
    * 
    * @param    schema
    *           The schema name.
    *           
    * @return   See above.
    */
   private static DataModelAst buildRootDataModel(String schema)
   {
      String pkg = String.format("%s.dmo", Configuration.getParameter("pkgroot"));
      
      // this is a reverse generation of what schema/p2o.xml does
      DataModelAst root = createDataModelAst(DataModelTokenTypes.DATA_MODEL, schema);
      root.putAnnotation(P2OLookup.PACKAGE, pkg);
      
      return root;
   }
   
   /**
    * Create a {@link DataModelAst data model object} AST for the specified buffer.
    * 
    * @param    buf
    *           The buffer instance.
    * @param    rbuf
    *           The {@link RecordBuffer} instance for the buffer.
    *           
    * @return   A {@link DataModelAst} buffer definition.
    * 
    * @throws   PersistenceException
    *           In case of problems during preparation of the temp-table schema and p2o ASTs.
    */
   private static DataModelAst buildDataModelClass(Buffer buf, RecordBuffer rbuf)
   throws PersistenceException
   {
      String historical = TableMapper.getLegacyName(buf);
      String dmoName    = rbuf.getDMOName();

      DataModelAst root = createDataModelAst(DataModelTokenTypes.CLASS, dmoName);
      root.putAnnotation(P2OLookup.HISTORICAL, historical);
      
      // build the ID property
      DataModelAst idProp = createDataModelAst(DataModelTokenTypes.PRIMARY, "id");
      idProp.putAnnotation(P2OLookup.DATATYPE, "Long");
      root.addChild(idProp);
      
      Map<String, DataModelAst> propMap = new HashMap<>();
      
      // go through all non-extent properties
      Class<?> dmoClass = rbuf.getDMOImplementationClass();
      Field[] fields = dmoClass.getDeclaredFields();
      for (Field field : fields)
      {
         if (!BaseDataType.class.isAssignableFrom(field.getType()))
            continue;
         
         DataModelAst propAst = buildProperty(buf, field, 0);
         
         root.addChild(propAst);
         
         propMap.put((String) propAst.getAnnotation("historical"), propAst);
      }
      
      // add the extent properties
      Class<?>[] innerClasses = dmoClass.getDeclaredClasses();
      for (Class<?> inner : innerClasses)
      {
         int extent = 0;
                  
         fields = inner.getDeclaredFields();
         for (Field field : fields)
         {
            if (!BaseDataType.class.isAssignableFrom(field.getType()))
               continue;
            
            if (extent == 0)
            {
               String fname = field.getName();
               
               String sizerMthdName = String.format("sizer%s%s",
                                                    Character.toUpperCase(fname.charAt(0)),
                                                    fname.substring(1));
               try
               {
                  Method sizerMthd = dmoClass.getMethod(sizerMthdName);
                  extent = ((integer) sizerMthd.invoke(buf)).toJavaIntegerType();
               }
               catch (IllegalAccessException    |
                      IllegalArgumentException  | 
                      InvocationTargetException | 
                      NoSuchMethodException     | 
                      SecurityException exc)
               {
                  String msg = "Error invoking the sizer method %s for property %s";
                  msg = String.format(msg, sizerMthdName, fname);

                  throw new PersistenceException(msg, exc);
               }
            }
            
            root.addChild(buildProperty(buf, field, extent));
         }
      }
      
      // composite info is not needed by P2OLookup
      
      Iterator<P2JIndex> idxIter = DMOIndex.databaseIndexes(rbuf.getSchema(),
                                                            rbuf.getDatabase(),
                                                            rbuf.getDMOInterface());
      while (idxIter.hasNext())
      {
         P2JIndex index = idxIter.next();
         String idx = index.getName();
         
         if (idx.startsWith("idx__"))
         {
            // eliminate the prefix
            idx = idx.substring("idx__".length());
         }
         
         String legacyName = TableMapper.getLegacyIndexName(buf, idx);
         if (legacyName != null)
         {
            // include only legacy indexes
            root.addChild(buildIndex(buf, rbuf, index, idx, propMap));
         }
      }
      
      // DataModelTokenTypes.UNIQUE, ONE_TO_ONE and ONE_TO_MANY nodes are not needed
      
      return root;
   }
   
   /**
    * Create a {@link DataModelAst property} AST for the specified DMO java field.
    * 
    * @param    buf
    *           The {@link Buffer} instance to which the property belongs.
    * @param    field
    *           The java field defining this property at the DMO.
    * @param    extent
    *           The legacy extent of this property.
    *           
    * @return   A {@link DataModelAst} property definition.
    */
   private static DataModelAst buildProperty(Buffer buf, Field field, int extent)
   {
      String property   = field.getName();
      String historical = TableMapper.getLegacyFieldName(buf, property);
      String datatype   = field.getType().getName();
      datatype = datatype.substring(datatype.lastIndexOf('.') + 1);
      
      DataModelAst root = createDataModelAst(DataModelTokenTypes.PROPERTY, property);
      root.putAnnotation(P2OLookup.HISTORICAL, historical);
      root.putAnnotation(P2OLookup.DATATYPE,   datatype);
      if (extent != 0)
      {
         root.putAnnotation(P2OLookup.EXTENT,     new Long(extent));
      }
      
      // INITIAL is not needed for P2OLookup
      
      return root;
   }
   
   /**
    * Build a {@link DataModelAst index} for the specified buffer.
    * 
    * @param    buf
    *           The {@link Buffer} instance to which the index belongs.
    * @param    rbuf
    *           The {@link RecordBuffer} instance.
    * @param    index
    *           The index definition.
    * @param    idxName
    *           The index name.
    * @param    propMap
    *           A map with the {@link DataModelAst property} ASTs by their legacy name.
    *           
    * @return   A {@link DataModelAst index} definition.
    */
   private static DataModelAst buildIndex(Buffer              buf,
                                          RecordBuffer        rbuf,
                                          P2JIndex            index,
                                          String              idxName,
                                          Map<String, DataModelAst> propMap)
   {
      String historical = TableMapper.getLegacyIndexName(buf, idxName);
      boolean unique    = index.isUnique();
      boolean primary   = index.isPrimary();
      boolean word      = index.isWord();
      
      DataModelAst root = createDataModelAst(DataModelTokenTypes.INDEX, idxName);
      root.putAnnotation(P2OLookup.HISTORICAL, historical);
      if (unique)
      {
         root.putAnnotation(P2OLookup.UNIQUE, unique);
      }
      if (primary)
      {
         root.putAnnotation(P2OLookup.PRIMARY, primary);
      }
      if (word)
      {
         root.putAnnotation(P2OLookup.WORD, word);
      }
      
      PersistentClass pclass = DatabaseManager.getClassMapping(rbuf.getDatabase(),
                                                               rbuf.getDMOImplementationClass());
      
      Map<String, String> propsByField = new HashMap<String, String>();
      Iterator<P2JIndexComponent> compIter = index.components();
      while (compIter.hasNext())
      {
         P2JIndexComponent comp = compIter.next();
         String fieldName = comp.getName();
         String propertyName = null;
         
         propertyName = propsByField.get(fieldName);
         if (propertyName == null)
         {
            Iterator<Property> itrProps = pclass.getDeclaredPropertyIterator();
            while (itrProps.hasNext())
            {
               Property hprop = itrProps.next();
               if (hprop.getColumnSpan() == 1)
               {
                  String field = ((Column) hprop.getColumnIterator().next()).getName();
                  if (field.equals(fieldName))
                  {
                     propertyName = hprop.getName();
                     break;
                  }
               }
               else
               {
                  // TODO: multi-col span properties
               }
            }
            
            propsByField.put(fieldName, propertyName);
         }
         
         String histName = TableMapper.getLegacyFieldName(buf, propertyName);
         if (histName != null)
         {
            // include only legacy fields
            root.addChild(buildIndexComponent(comp, propMap.get(histName)));
         }
      }
      
      return root;
   }
   
   /**
    * Build an {@link DataModelAst index component} associated with the given property AST.
    * 
    * @param    comp
    *           The index component 
    * @param    propAst
    *           The associated {@link DataModelAst property AST}.
    *           
    * @return   See above.
    */
   private static DataModelAst buildIndexComponent(P2JIndexComponent comp,
                                                   DataModelAst      propAst)
   {
      String name        = comp.getName();
      String historical  = (String) propAst.getAnnotation(P2OLookup.HISTORICAL);
      String datatype    = (String) propAst.getAnnotation(P2OLookup.DATATYPE);

      DataModelAst root = createDataModelAst(DataModelTokenTypes.INDEX_COL, name);
      root.putAnnotation(P2OLookup.HISTORICAL,  historical);
      root.putAnnotation(P2OLookup.DATATYPE,    datatype);
      if (comp.isIgnoreCase())
      {
         root.putAnnotation(P2OLookup.IGNORE_CASE, comp.isIgnoreCase());
      }
      if (comp.isDescending())
      {
         root.putAnnotation(P2OLookup.DESCEND,     comp.isDescending());
      }
      if (comp.isAbbreviated())
      {
         root.putAnnotation(P2OLookup.ABBREVIATED, comp.isAbbreviated());
      }
      root.putAnnotation(P2OLookup.REFID,       propAst.getId());
      
      return root;
   }
   
   /**
    * Create a {@link DataModelAst} with the given token type and text.
    * 
    * @param    type
    *           The node's token type.
    * @param    text
    *           The node's text.
    *           
    * @return   A {@link DataModelAst} instance setup as noted above.
    */
   private static DataModelAst createDataModelAst(int type, String text)
   {
      long treeId = AstManager.get().getTreeId(locate().p2oFile);
      DataModelAst ast = new DataModelAst(new CommonToken(type, text));
      ast.setId(AstManager.get().getNextNodeId(treeId));
      return ast;
   }
   
   /**
    * Create a {@link ProgressAst} with the given token type and text.
    * 
    * @param    type
    *           The node's token type.
    * @param    text
    *           The node's text.
    *           
    * @return   A {@link ProgressAst} instance setup as noted above.
    */
   private static ProgressAst createProgressAst(int type, String text)
   {
      return new ProgressAst(new CommonToken(type, text));
   }
   
   /**
    * Locate the context-local {@link WorkArea} instance.
    * 
    * @return   See above.
    */
   private static WorkArea locate()
   {
      return local.get();
   }

   /**
    * Shows a conversion error in a modal dialog and registers it in the error message list; does
    * not set the ERROR-STATUS:ERROR flag. The format of the message is:
    * <pre>
    * Could not [details] the '{@link WorkArea#predicate}' predicate for query {@link WorkArea#qname}
    * </pre>
    * 
    * @param    exc
    *           The exception causing the conversion error.
    * @param    details
    *           Additional details about this error, to be shown in the message.
    */
   private static void showConversionError(Exception exc, String details)
   {
      WorkArea wa = locate();
      
      String msg = "Could not %s the '%s' predicate for query %s";
      msg = String.format(msg, details, wa.predicate, wa.qname);
      
      ErrorManager.recordOrShowError(-1, msg, true, false, false);
      
      if (LOG.isLoggable(Level.SEVERE))
      {
         LOG.log(Level.SEVERE, msg, exc);
      }
   }
   
   /**
    * Container of context-local data.
    * <p>
    * Implements {@link Finalizable} to unregister any surviving dynamic queries from the 
    * class-loader.
    */
   private static class WorkArea
   implements Finalizable
   {
      /**
       * Map of active in-mem compiled query classes, per their child P2JQuery instance; 
       * the values are not P2JQuery subclasses, they are the java classes compiled from a
       * converted 4GL query snippet.
       */
      private final Map<P2JQuery, Class<?>> dynamicQueries = new HashMap<P2JQuery, Class<?>>();

      /** Flag indicating that the permanent schemas have been loaded for this context. */
      private boolean permSchemasLoaded = false;

      // temporary data about the currently converted query
      
      /** The query predicate being converted. */
      private String predicate = null;
      
      /** The name of the QUERY resource. */
      private String qname = null;
      
      /** The buffers used by the QUERY resource. */
      private List<Buffer> buffers = null;
      
      /** The name of the class containing the converted query code. */
      private String dynQueryClass = null;
      
      /** The full name of the class containing the converted query code. */
      private String fullClassName = null;
      
      /** The file name of the class containing the converted query code. */
      private String dynQueryFile = null;
      
      /** The name of the generated {@link JavaAst} tree. */
      private String jastFile = null;
      
      /** The name of the converted {@link ProgressAst} tree. */
      private String astFile = null;
      
      /** The name of the {@link DataModelAst} tree with the temp-table objects for this query. */
      private String p2oFile = null;

      /**
       * Called when the user's session ends; it unregisters from the class-loader any surviving
       * converted query class.
       */
      @Override
      public void finished()
      {
         Set<P2JQuery> keys = new HashSet<>(dynamicQueries.keySet());
         for (P2JQuery query : keys)
         {
            unregisterQuery(query);
         }
      }

      /** No-op. */
      public void iterate()
      {
      }

      /** No-op. */
      public void retry()
      {
      }
   }
}
