/*
** Module   : TempTable.java
** Abstract : Declaration of some dynamic TEMP-TABLES specific features (attributes, methods).
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 OM  20130129 Created initial version.
** 002 CA  20130221 Added ADM-DATA and UNIQUE-ID support.
** 003 OM  20130301 Pushed up defaultBufferHandle() to Buffer interface.
** 004 CA  20130311 Added missing ADD-NEW-FIELD and ADD-NEW-INDEX API versions, using wrapper 
**                  types. Note that the conversion will use only the APIs with the wrapper type
**                  parameters; the other APIs (using non-wrapper parameters) are removed.
** 005 SVL 20130822 Removed uniqueId. Added overloaded methods for several functions.
** 006 CA  20130927 Added LegacyResource annotation, to determine resource type.
** 007 EVL 20131002 Adding legacy attributes and methods annotation.
** 008 EVL 20131010 Adding Errorable interface to support ERROR legacy attribute.
** 009 CA  20131021 Adds DynamicResource interface to support DYNAMIC attribute.
** 010 SVL 20131115 Added getDMOClass.
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;

/**
 * Declaration of dynamic TEMP-TABLES attributes and methods.
 * This interface is needed by {@link CommonHandle} in order to store 
 * the handle to a temp table.
 * 
 * @author om
 */
@LegacyResource(resource = LegacyResource.TEMP_TABLE)
public interface TempTable
extends BufferHandle,
        ADMData,
        Errorable,
        UniqueID,
        DynamicResource
{
   /**
    * Duplicates the structure of an existing table, copying the field 
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the definitions.
    *
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "CREATE-LIKE")
   public logical createTableLike(handle hbuf);

   /**
    * Duplicates the structure of an existing table, copying the field 
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   tableName
    *          The name of a table to create the copy of.
    *
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "CREATE-LIKE")
   public logical createTableLike(String tableName);

   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   tableName
    *          The name of a table to create the copy of.
    *
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "CREATE-LIKE")
   public logical createTableLike(character tableName);

   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the definitions.
    * @param   indexName
    *          The name of a single index which is copied from the source table.
    *
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "CREATE-LIKE")
   public logical createTableLike(handle hbuf, String indexName);

   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the definitions.
    * @param   indexName
    *          The name of a single index which is copied from the source table.
    *
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "CREATE-LIKE")
   public logical createTableLike(handle hbuf, character indexName);


   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   tableName
    *          The name of a table to create the copy of.
    * @param   indexName
    *          The name of a single index which is copied from the source table.
    *
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "CREATE-LIKE")
   public logical createTableLike(String tableName, String indexName);

   /**
    * Duplicates the structure of an existing table, copying the field
    * definitions, and optionally, the indexes.
    * This method is the P2J equivalent of <code>CREATE-LIKE</code> method
    * of Progress 4GL.
    *
    * @param   tableName
    *          The name of a table to create the copy of.
    * @param   indexName
    *          The name of a single index which is copied from the source table.
    *
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "CREATE-LIKE")
   public logical createTableLike(character tableName, character indexName);

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    *          
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-NEW-FIELD")
   public logical addNewField(character name, character type);

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    *          
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-NEW-FIELD")
   public logical addNewField(character name, character type, integer extent);

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    *          
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-NEW-FIELD")
   public logical addNewField(character name, character type, integer extent, character format);

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    *          
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-NEW-FIELD")
   public logical addNewField(character    name,
                              character    type,
                              integer      extent,
                              character    format,
                              BaseDataType initial);

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    * @param   label
    *          The label of the defined field. If null or unknown the name 
    *          parameter will be used.
    *          
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-NEW-FIELD")
   public logical addNewField(character    name,
                              character    type,
                              integer      extent,
                              character    format,
                              BaseDataType initial,
                              character    label);

   /**
    * Adds a field with the specified properties to the temp-table.
    * This method is the P2J equivalent of <code>ADD-NEW-FIELD</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   type
    *          The data type of the specified field.
    * @param   extent
    *          An integer expression specifying the extent of an array.
    * @param   format
    *          The data format for the defined data type. If null or unknown, 
    *          the default format will be used.
    * @param   initial
    *          An expression that evaluates to the initial value of the 
    *          defined field. 
    *          TODO: this method will probably be overloaded because of this.
    * @param   label
    *          The label of the defined field. If null or unknown the name 
    *          parameter will be used.
    * @param   columnLabel
    *          The label of the column associated with the defined field
    *          
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-NEW-FIELD")
   public logical addNewField(character    name,
                              character    type,
                              integer      extent,
                              character    format,
                              BaseDataType initial,
                              character    label,
                              character    columnLabel);

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * 
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-NEW-INDEX")
   public logical addNewIndex(character name);

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * 
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-NEW-INDEX")
   public logical addNewIndex(character name, logical unique);

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * @param   primary
    *          TRUE if this is the primary index.
    * 
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-NEW-INDEX")
   public logical addNewIndex(character name, logical unique, logical primary);

   /**
    * Adds a new empty index with the specified name to the temp-table. Index 
    * components must be added with the <code>ADD-INDEX-FIELD()</code> method.
    * This method is the P2J equivalent of <code>ADD-NEW-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to be created.
    * @param   unique
    *          TRUE if this index is unique.
    * @param   primary
    *          TRUE if this is the primary index.
    * @param   word
    *          TRUE if this is a word index.
    * 
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-NEW-INDEX")
   public logical addNewIndex(character name, logical unique, logical primary, logical word);

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Database field name from which to copy the field. The table 
    *          name must be qualified with the database name.
    * 
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-LIKE-FIELD")
   public logical addFieldLike(String name, String source);

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Database field name from which to copy the field. The table 
    *          name must be qualified with the database name.
    * 
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-LIKE-FIELD")
   public logical addFieldLike(character name, character source);

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method of Progress 4GL.
    *
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Handle to field name from which to copy the field.
    *
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-LIKE-FIELD")
   public logical addFieldLike(String name, handle source);

   /**
    * Adds to the temp-table a field like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-FIELD</code> method of Progress 4GL.
    * 
    * @param   name
    *          The name of the field to be created in the temp-table.
    * @param   source
    *          Handle to field name from which to copy the field.
    * 
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-LIKE-FIELD")
   public logical addFieldLike(character name, handle source);

   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-INDEX</code> method of Progress 4GL.
    * 
    * @param   name
    *          The name of the index to which the source index is being copied. Progress ignores
    *          this option and uses the name of the source index instead.
    * @param   sourceName
    *          The name of the index in the source table that is being copied 
    *          to the temp-table.
    * @param   dbTable
    *          The database table name from which to copy the index.
    * 
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-LIKE-INDEX")
   public logical addIndexLike(String name, String sourceName, String dbTable);

   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-INDEX</code> method
    * of Progress 4GL.
    * 
    * @param   name
    *          The name of the index to which the source index is being copied. Progress ignores
    *          this option and uses the name of the source index instead.
    * @param   sourceName
    *          The name of the index in the source table that is being copied 
    *          to the temp-table.
    * @param   dbTable
    *          The database table name from which to copy the index.
    * 
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-LIKE-INDEX")
   public logical addIndexLike(character name, character sourceName, character dbTable);

   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to which the source index is being copied. Progress ignores
    *          this option and uses the name of the source index instead.
    * @param   sourceName
    *          The name of the index in the source table that is being copied
    *          to the temp-table.
    * @param   sourceBuffer
    *          Buffer handle from which to copy the index.
    *
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-LIKE-INDEX")
   public logical addIndexLike(String name, String sourceName, handle sourceBuffer);

   /**
    * Adds to the temp-table an index like the one specified.
    * This method is the P2J equivalent of <code>ADD-LIKE-INDEX</code> method
    * of Progress 4GL.
    *
    * @param   name
    *          The name of the index to which the source index is being copied. Progress ignores
    *          this option and uses the name of the source index instead.
    * @param   sourceName
    *          The name of the index in the source table that is being copied
    *          to the temp-table.
    * @param   sourceBuffer
    *          Buffer handle from which to copy the index.
    *
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "ADD-LIKE-INDEX")
   public logical addIndexLike(character name, character sourceName, handle sourceBuffer);

   /**
    * Signals that all the field and index definitions for a temp-table 
    * have been provided. After the call to this method no fields and indexes
    * can be added to this temporary table.
    * This method is the P2J equivalent of <code>TEMP-TABLE-PREPARE</code> 
    * method of Progress 4GL.
    * 
    * @param   name
    *          Temp-table name to be used in subsequent query statements that refer to 
    *          this temp-table.
    *
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "TEMP-TABLE-PREPARE")
   public logical tempTablePrepare(String name);

   /**
    * Signals that all the field and index definitions for a temp-table have been provided. After
    * the call to this method no fields and indexes can be added to this temporary table.
    * This method is the P2J equivalent of <code>TEMP-TABLE-PREPARE</code> 
    * method of Progress 4GL.
    * 
    * @param   name
    *          Temp-table name to be used in subsequent query statements that
    *          refer to this temp-table.
    *          
    * @return  <code>true</code> on success.
    */
   @LegacyMethod(name = "TEMP-TABLE-PREPARE")
   public logical tempTablePrepare(character name);

   /**
    * Returns <code>true</code> if this temp-table is in prepared state (after a successful call 
    * to {@link #tempTablePrepare}). 
    * This method implements the read-only PREPARED attribute from 4GL.
    * 
    * @return  true if operation is successful 
    */
   @LegacyAttribute(name = "PREPARED")
   public logical prepared();

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    *
    * @return  true if operation is successful
    */
   @LegacyMethod(name = "COPY-TEMP-TABLE")
   public logical copyTempTable(handle other);

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    *
    * @return  true if operation is successful
    */
   @LegacyMethod(name = "COPY-TEMP-TABLE")
   public logical copyTempTable(handle other, logical append);

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    *
    * @return  true if operation is successful
    */
   @LegacyMethod(name = "COPY-TEMP-TABLE")
   public logical copyTempTable(handle other, logical append, logical replace);

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    *
    * @return  true if operation is successful
    */
   @LegacyMethod(name = "COPY-TEMP-TABLE")
   public logical copyTempTable(handle other, logical append, logical replace, logical looseCopy);

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    *
    * @return  true if operation is successful
    */
   @LegacyMethod(name = "COPY-TEMP-TABLE")
   public logical copyTempTable(handle other, boolean append);

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    *
    * @return  true if operation is successful
    */
   @LegacyMethod(name = "COPY-TEMP-TABLE")
   public logical copyTempTable(handle other, boolean append, boolean replace);

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current
    * records are dropped before copy operation.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode.
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    *
    * @return  true if operation is successful
    */
   @LegacyMethod(name = "COPY-TEMP-TABLE")
   public logical copyTempTable(handle other, boolean append, boolean replace, boolean looseCopy);

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current 
    * records are dropped before copy operation. 
    * Only the first parameter is mandatory so this method will probably be heavy overloaded.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode. 
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    * @param   prefix
    *          Prefix for naming the copied temp-table.
    *
    * @return  true if operation is successful 
    */
   @LegacyMethod(name = "COPY-TEMP-TABLE")
   public logical copyTempTable(handle    other,
                                logical   append,
                                logical   replace,
                                logical   looseCopy,
                                character prefix);

   /**
    * Copies the other temp-table. other must be a temp-tables handle By default, all current 
    * records are dropped before copy operation. 
    * Only the first parameter is mandatory so this method will probably be heavy overloaded.
    *
    * @param   other
    *          Some other temp-table to copy from.
    * @param   append
    *          Append mode. 
    * @param   replace
    *          Replace mode.
    * @param   looseCopy
    *          Loose copy mode.
    * @param   prefix
    *          Prefix for naming the copied temp-table.
    *
    * @return  true if operation is successful 
    */
   @LegacyMethod(name = "COPY-TEMP-TABLE")
   public logical copyTempTable(handle  other,
                                boolean append,
                                boolean replace,
                                boolean looseCopy,
                                String  prefix);
   /**
    * Sets the UNDO attribute of 4GL temp-table.
    *
    * @param   undo
    *          The new value for the attribute.
    */
   @LegacyAttribute(name = "UNDO", setter = true)
   public void setCanUndo(logical undo);

   /** 
    * Obtain the current UNDO attribute of 4GL temp-table.
    * 
    * @return  true if the the temp-table can undo.
    */
   @LegacyAttribute(name = "UNDO")
   public logical canUndo();

   /**
    * Sets the ERROR-STRING attribute of 4GL temp-table.
    *
    * @param   errStr
    *          The new value for the ERROR-STRING attribute.
    */
   @LegacyAttribute(name = "ERROR-STRING", setter = true)
   public void changeErrorString(character errStr);

   /** 
    * Obtain the current ERROR-STRING attribute of 4GL temp-table.
    * 
    * @return  the current ERROR-STRING of the temp-table
    */
   @LegacyAttribute(name = "ERROR-STRING")
   public character errorString();
   
   /**
    * Checks whether this temp-table contains any records.
    * 
    * @return  true if this temp-table has records.
    */
   @LegacyAttribute(name = "HAS-RECORDS")
   public logical hasRecords();

   /**
    * Returns the temp-table to it's clear state, removing all defined fields and indexes.
    * 
    * TODO: At the moment of writing this it's evident that the data from the table should also be
    * removed or method should fail (return false) if the table contains any data.
    * 
    * @return  true if operation is successful
    */
   @LegacyMethod(name = "CLEAR")
   public logical clear();

   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-INDEX-FIELD")
   public logical addFieldToIndex(String indexName, String field, String ascending);
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-INDEX-FIELD")
   public logical addFieldToIndex(String indexName, String field, character ascending);
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-INDEX-FIELD")
   public logical addFieldToIndex(character indexName, String field, String ascending);
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-INDEX-FIELD")
   public logical addFieldToIndex(character indexName, String field, character ascending);
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-INDEX-FIELD")
   public logical addFieldToIndex(String indexName, character field, String ascending);
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-INDEX-FIELD")
   public logical addFieldToIndex(String indexName, character field, character ascending);
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-INDEX-FIELD")
   public logical addFieldToIndex(character indexName, character field, String ascending);
   
   /**
    * Construct an index of this temp-table by adding a new field to it. This method requires
    * the index to be first declared with {@link #addNewIndex} and cannot be called after the 
    * {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character the first 2 
    * and logical the last), this method will eventually need to be overloaded with those 
    * signatures.
    * 
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    * @param   ascending
    *          "asc" if the ascending ordering, "desc" for descending other -> error.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-INDEX-FIELD")
   public logical addFieldToIndex(character indexName, character field, character ascending);

   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-INDEX-FIELD")
   public logical addFieldToIndex(String indexName, String field);
   
   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   @LegacyMethod(name = "ADD-INDEX-FIELD")
   public logical addFieldToIndex(character indexName, String field);
   
   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   @LegacyMethod(name = "ADD-INDEX-FIELD")
   public logical addFieldToIndex(String indexName, character field);
   
   /**
    * Construct an index of this temp-table by adding a new field to it. The new added filed will
    * sort ascending the table. This method requires the index to be first declared with 
    * {@link #addNewIndex} and cannot be called after the {@link #prepared()}.
    * 
    * TODO: as the parameters of this methods are all expressions (of type character), this 
    * method will eventually need to be overloaded with those signatures.
    *
    * @param   indexName
    *          The name of the index being built.
    * @param   field
    *          The name of the temp-table to be added to the index.
    *
    * @return  <code>true</code> for success.
    */
   
   @LegacyMethod(name = "ADD-INDEX-FIELD")
   public logical addFieldToIndex(character indexName, character field);

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL. 
    * 
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the field definitions.
    * 
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-FIELDS-FROM")
   public logical addFieldsFrom(handle hbuf);

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL. 
    * 
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the field definitions.
    * @param   except
    *          A comma-separated list of field names to be excluded from the copy process.
    * 
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-FIELDS-FROM")
   public logical addFieldsFrom(handle hbuf, String except);

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL.
    *
    * @param   hbuf
    *          A handle to a buffer or table from which to copy the field definitions.
    * @param   except
    *          A comma-separated list of field names to be excluded from the copy process.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-FIELDS-FROM")
   public logical addFieldsFrom(handle hbuf, character except);

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL. 
    * 
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    * 
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-FIELDS-FROM")
   public logical addFieldsFrom(String tableName);

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL.
    *
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-FIELDS-FROM")
   public logical addFieldsFrom(character tableName);

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL.
    *
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    * @param   except
    *          A comma-separated list of field names to be excluded from the copy process.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-FIELDS-FROM")
   public logical addFieldsFrom(String tableName, character except);

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL.
    *
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    * @param   except
    *          A comma-separated list of field names to be excluded from the copy process.
    *
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-FIELDS-FROM")
   public logical addFieldsFrom(character tableName, character except);

   /**
    * Adds fields from another table into this temp-table. The fields that are already in the
    * temp-table are ignored.
    * This is the P2J implementation of the ADD-FIELDS-FROM of 4GL. 
    * 
    * @param   tableName
    *          The name of the table from which to copy the field definitions.
    * @param   except
    *          A comma-separated list of field names to be excluded from the copy process.
    * 
    * @return  <code>true</code> for success.
    */
   @LegacyMethod(name = "ADD-FIELDS-FROM")
   public logical addFieldsFrom(String tableName, String except);

   /**
    * Get DMO implementation class associated with this table.
    *
    * @return DMO implementation class associated with this table or <code>null</code> if the
    *         table is not initialized yet.
    */
   public Class getDMOClass();
}
